//Black Deck disc image generator v1.0
//This work is placed in the public domain

#include <cstdlib>
#include <iostream>
#include <fstream>
#include "pngwriter-0.5.4/src/pngwriter.h"
#include "MersenneTwister.h"

using namespace std;

const double pi = 2*acos(0.0);
MTRand rng;

class PatternGenerator {
public:
    virtual float get(double angle, double distance) const = 0; // angle = 0..<2*pi, distance = 0..<1
};

class SampleSpiral : public PatternGenerator {
private:
    double samplesPerCycle;
    double grooves;
    int sampleSize;
    unsigned char *sample;
public:
    float get(double angle, double distance) const {
        double phase = angle / (2*pi);
        int groove = (int) floor(grooves*distance-phase + 1);
        double accurateSamplePos = (groove+phase)*samplesPerCycle;
        int samplePos = (int)floor(accurateSamplePos);
        float r;
        if (samplePos < sampleSize) {
            double fractPos = accurateSamplePos-samplePos;
            double ym1 = sample[samplePos-1];
            double y0 = sample[samplePos];
            double y1 = sample[samplePos+1];
            double y2 = sample[samplePos+2];
            float c0 = y0;
            float c1 = 1/2.0*(y1-ym1);
            float c2 = ym1 - 5/2.0*y0 + 2*y1 - 1/2.0*y2;
            float c3 = 1/2.0*(y2-ym1) + 3/2.0*(y0-y1);
            r = ((c3*fractPos+c2)*fractPos+c1)*fractPos+c0;
            r *= (1.0/255);
        } else {
            r = 0.5;
        }
        return r;
    }
    SampleSpiral(const char *fileName, double samplingFrequency, double rotationFrequency) {
        samplesPerCycle = samplingFrequency / rotationFrequency;
        grooves = sampleSize / samplesPerCycle - 2;
        ifstream file(fileName, ios::in|ios::binary|ios::ate);
        sampleSize = file.tellg();
        sample = new unsigned char [sampleSize+20];
        for (int t = 0; t < 10; t++) {
           sample[t] = 0;
           sample[sampleSize+10+t] = 0;
        }
        sample += 10;
        file.seekg(0, ios::beg);
        file.read ((char  *)sample, sampleSize);
        file.close();
    }
    
    ~SampleSpiral() {
        sample -= 10;
        delete[] sample;
    }
};

double getNoteFreq(double baseFreq, int note) {
   return baseFreq * pow(2, note/12.0);
}

class TrackAngularPeriods {
public:
    virtual double get(int trackIndex) const = 0;
    virtual int getNumTracks() const = 0;
};

class NoteAngularPeriods : public TrackAngularPeriods {
private:
    double *angularPeriods;
    int numNotes;
public:
    int getNumTracks() const {
        return numNotes;
    }
    double get(int noteIndex) const {
        return angularPeriods[noteIndex];
    }
    NoteAngularPeriods(int numNotes, double rotationFreq, double baseFreq, const int *notes) {
        this->numNotes = numNotes;
        angularPeriods = new double[numNotes];
        for (int t = 0; t < numNotes; t++) {
            double freq = getNoteFreq(baseFreq, notes[t]);
            double cycles = round(freq/rotationFreq);
            angularPeriods[t] = 2*pi/cycles;
            // double error = log(cycles/freq/rotationFreq)/log(2.0)*12;
            // printf("%f->%f (err %f semitones)\n", freq, cycles*rotationFreq, error);
        }
    }
    ~NoteAngularPeriods() {
        delete[] angularPeriods;
    }
};

class NoteHarmonicAngularPeriods : public TrackAngularPeriods {
private:
    double *angularPeriods;
    int numNotes;
    int numHarmonics;
public:
    int getNumTracks() const {
        return numNotes*numHarmonics;
    }
    double get(int noteIndex) const {
        return angularPeriods[noteIndex];
    }
    NoteHarmonicAngularPeriods(int numNotes, double rotationFreq, double baseFreq, const int *notes, int numHarmonics) {
        this->numNotes = numNotes;
        this->numHarmonics = numHarmonics;
        angularPeriods = new double[numNotes*numHarmonics];
        for (int t = 0; t < numNotes; t++) {
            for (int u = 0; u < numHarmonics; u++) {
                double freq = getNoteFreq(baseFreq, notes[t])*(u+1);
                double cycles = round(freq/rotationFreq);
                angularPeriods[t*numHarmonics + u] = 2*pi/cycles;
                // double error = log(cycles/freq/rotationFreq)/log(2.0)*12;
                // printf("%f->%f (err %f semitones)\n", freq, cycles*rotationFreq, error);
            }
        }
    }
    ~NoteHarmonicAngularPeriods() {
        delete[] angularPeriods;
    }
};

class NoteOctaveAngularPeriods : public TrackAngularPeriods {
private:
    double *angularPeriods;
    int numNotes;
    int numOctaves;
public:
    int getNumTracks() const {
        return numNotes*numOctaves;
    }
    double get(int noteIndex) const {
        return angularPeriods[noteIndex];
    }
    NoteOctaveAngularPeriods(int numNotes, double rotationFreq, double baseFreq, const int *notes, int numOctaves) {
        this->numNotes = numNotes;
        this->numOctaves = numOctaves;
        angularPeriods = new double[numNotes*numOctaves];
        for (int t = 0; t < numNotes; t++) {
            for (int u = 0; u < numOctaves; u++) {
                double freq = getNoteFreq(baseFreq, notes[t]+12*u);
                double cycles = round(freq/rotationFreq);
                angularPeriods[t*numOctaves + u] = 2*pi/cycles;
                // double error = log(cycles/freq/rotationFreq)/log(2.0)*12;
                // printf("%f->%f (err %f semitones)\n", freq, cycles*rotationFreq, error);
            }
        }
    }
    ~NoteOctaveAngularPeriods() {
        delete[] angularPeriods;
    }
};

class WaveformGenerator {
public:
    virtual float get(double phase, double tone) const = 0; // phase = 0..1, tone = 0..1
};

class Pulsewave : public WaveformGenerator {
public:
    float get(double phase, double tone) const {
        if (tone < 0.1) {
            tone = 0;
        } else {
            tone = (tone-0.1)*(1.0/0.9);
        }
        double threshold = (tone-1)*(tone-1)*0.25;
        if (phase < threshold || phase > 1-threshold) {
            return 0;
        } else {
            return 1;
        }
    }
};

class Sinewave : public WaveformGenerator {
public:
    float get(double phase, double tone) const {
        return sin(2*pi*phase)*0.5 + 0.5;
    }
};

class Sawwave : public WaveformGenerator {
public:
    float get(double phase, double tone) const {
        return phase;
    }
};

class Squarewave : public WaveformGenerator {
public:
    float get(double phase, double tone) const {
        if (phase > 0.5) {
            return 1;
        } else {
            return 0;
        }
    }
};

class WaveformNotes : public PatternGenerator {
private:
    TrackAngularPeriods const *periods;
    WaveformGenerator const *waveform;
    float *displacements;
public:
    float get(double angle, double distance) const {
        double temp = periods->getNumTracks();
        int noteIndex = (int)floor(distance*periods->getNumTracks());
        double phase = fmod(angle + displacements[noteIndex], periods->get(noteIndex))/periods->get(noteIndex);
        double tone = distance*periods->getNumTracks()-noteIndex;
        return waveform->get(phase, tone);
    }
    WaveformNotes(TrackAngularPeriods const *periods, WaveformGenerator const *waveform) {
        this->periods = periods;
        this->waveform = waveform;
        displacements = new float[periods->getNumTracks()];
        for (int t = 0; t < periods->getNumTracks(); t++) {
            displacements[t] = rng.randExc()*2*pi;
        }
    }
    WaveformNotes() {
        delete[] displacements;
    }
};

void minor(int *notes, int base, int start, int numNotes) {
    const int minorNotes[7] = {0, 2, 3, 5, 7, 8, 10};
    while (start < 0) {
        start = start + 666*7;
        base = base - 666*12;
    }
    int index = start % 7;
    base += 12 * (start / 7);
    for (int t = 0; t < numNotes; t++) {
        notes[t] = base + minorNotes[index];
        index++;
        if (index >= 7) {
            base += 12;
            index = 0;
        }
    }
}

int main(int argc, char *argv[])
{
    // Image settings. Edit these
    const char imageFileName[] = "disc.png"; // Image file name, default: "disc.png"
    const double dpi = 75; // Printer dots per inch (dpi) in pixels/inch, final: 1200, draft: 75
    const int antialias = 1; // Square root of the number of antialias samples per output pixel, final: 8, draft: 1
    const int margin = 2; // Image margins in pixels, default: 2

    // Printer calibration. Edit these
    const double xCorrection = 268.05/268; // Printer correction factor for x axis, default: 1
    const double yCorrection = 268.6/268; // Printer correction factor for y axis, default: 1
    const double gamma = 1.169; // Gamma adjustment, default: 1.1

    // Tuning. Edit this
    const double tuningFreq = 440; // Tuning frequency for note generation in Hz, default: 440

    // Physical and mechanical quantities. Edit these only if you have changed the disc "standard" format
    const int numTracks = 8; // Number of tracks, default: 8
    const double rotationFreq = 3.5198; // Rotation frequency in Hz, default value from 2006: 3.5198, 2010 value: 3.43858853, for 125 BPM: 2.08333..., compromise value for 440 Hz tuning: 3.6872
    const double shaftRadius = 3.5; // Radius of shaft in mm, default: 3.5
    const double crossHairRadius = 10; // Radius of the middle crosshair in mm, default: 10
    const double lineWidth = 1; // Width of all line art in mm, default: 1
    const double innerRadius = 50; // Distance from shaft center to outer edge of center graphics area in mm, default: 50
    const double outerRadius = 142.5; // Distance from shaft center to outer edge of disc boundary line in mm, default: 142.5
    const double trackSpacing = 80.0/numTracks; // Spacing of tracks and width of each track in mm, default: 10
    const double track0InnerRadius = 54; // Distance from shaft center to inner edge of in-most track in mm, default: 54
    const double motorMinRadius = 14; // (Currently unused) Distance from shaft center to edge of base plate in mm, default: 14
    const double motorMaxRadius = 59; // (Currently unused) Distance from shaft center to edge of tonearm in mm, default: 59

    // Derived quantities. Do not edit
    const double dpmm = dpi/25.4; // pixels/mm
    const double sdpmm = dpmm*antialias; // pixels/mm
    const double trackMaxOuterRadius = track0InnerRadius + numTracks*trackSpacing;
    const double squareTrack0InnerRadius = track0InnerRadius*track0InnerRadius;
    const double squareTrack0InnerLineRadius = (track0InnerRadius-lineWidth)*(track0InnerRadius-lineWidth);
    const double squareTrackMaxOuterRadius = trackMaxOuterRadius*trackMaxOuterRadius;
    const double squareTrackMaxOuterLineRadius = (trackMaxOuterRadius+lineWidth)*(trackMaxOuterRadius+lineWidth);
    const double squareInnerRadius = innerRadius*innerRadius;
    const double squareOuterRadius = outerRadius*outerRadius;
    const double squareOuterLineRadius = (outerRadius-lineWidth)*(outerRadius-lineWidth);
    const double squareShaftRadius = shaftRadius*shaftRadius;
    const double squareCrossHairRadius = crossHairRadius*crossHairRadius;
    const double squareMotorMinRadius = motorMinRadius*motorMinRadius;
    const double squareMotorMaxRadius = motorMaxRadius*motorMaxRadius;
    const double invSquareAntialias = 1.0/(antialias*antialias);
    const int imageWidthInPixels = (int)(dpmm*outerRadius*2)+margin*2;
    const int imageWidthInSubpixels = imageWidthInPixels*antialias;

    // Pattern generator. Edit this
    //SampleSpiral generator("chant.pcm", 44100, rotationFreq);
    int notes[numTracks];
    minor(notes, -24, 0, numTracks);
//    NoteHarmonicAngularPeriods periods(numTracks, rotationFreq, tuningFreq, notes, 5);
    NoteAngularPeriods periods(numTracks, rotationFreq, tuningFreq, notes);
    Sawwave waveform;
    WaveformNotes generator(&periods, &waveform);

    // Disc image generation. No need to touch.
    pngwriter png(imageWidthInPixels, imageWidthInPixels, 0, imageFileName);
    int lastPercent = -1;
    for (int pixelY = 0; pixelY < imageWidthInPixels; pixelY++) {
        int percent = 100*pixelY/(imageWidthInPixels-1);
        if (lastPercent != percent) {
            lastPercent = percent;
            printf("\r%d%%", percent);
        }
        for (int pixelX = 0; pixelX < imageWidthInPixels; pixelX++) {
            double totalR = 0;
            for (int subPixely = pixelY*antialias; subPixely < pixelY*antialias+antialias; subPixely++) {
                double y = (subPixely-imageWidthInSubpixels/2.0)/sdpmm;
                y *= yCorrection;
                for (int subPixelx = pixelX*antialias; subPixelx < pixelX*antialias+antialias; subPixelx++) {
                    double x = (subPixelx-imageWidthInSubpixels/2.0)/sdpmm;
                    x *= xCorrection;
                    double squareRadius = x*x+y*y;
                    double r = 1;//, g = 1, b = 1;
                    if (squareRadius > squareOuterRadius) {
                        r = 1;
                    } else if (squareRadius >= squareTrackMaxOuterRadius) {
                        if (squareRadius < squareTrackMaxOuterLineRadius) {
                            r = 0;
                        } else if (squareRadius > squareOuterLineRadius) {
                            r = 0;
                        } else {
                            r = 1;
                        }
                    } else if (squareRadius < squareTrack0InnerRadius) {
                        if (squareRadius > squareTrack0InnerLineRadius) {
                            r = 0;
                        } else if (squareRadius > squareInnerRadius) {
                            r = 1;
                        } else if ((squareRadius < squareCrossHairRadius) && (fabs(x) < lineWidth/2 || fabs(y) < lineWidth/2)) {
                            r = 0;
                        } else if (squareRadius < squareShaftRadius) {
                            r = 0;
                        } else {
                            // Graphics, if any
                            r = 1;
                        }
                    } else {
                        double radius = sqrt(squareRadius);
                        double phase = pi + atan2(x, y);
                          float distance = (radius-track0InnerRadius)/(trackMaxOuterRadius-track0InnerRadius);
                          r = generator.get(phase, distance);
                    }
                    if (antialias > 1) {
                        r *= r;
                    }                                         
                    totalR += r;
                }
            }
            if (antialias > 1) {
              totalR *= invSquareAntialias;
              totalR = pow(totalR, 0.5*gamma);
            } else {
              totalR = pow(totalR, gamma);
            }
            png.plot(1+pixelX, 1+pixelY, totalR, totalR, totalR);
        }
    }
    png.close();
//    system("PAUSE");
    return EXIT_SUCCESS;
}
